/***********************************************************/
/*                  Graph Paritioning                      */
/*                      adjlist.c                          */
/*                                                         */
/*                Author:  James P. Fox                    */
/*                Advisor: Mark Goldberg                   */
/*                                                         */
/*                    Masters Project                      */
/*            Rensselaer Polytechnic Institute             */
/*                  Summer, Fall, 1990                     */
/***********************************************************/
/*     This module provides the routines for creation and  */
/*  operation on the random graphs needed to test the      */
/*  graph partitioning algorithms.  Routines create the    */
/*  adjacency lists and compute the "degree" of each vertex*/
/*  in the graph.  A routine is also provided to print out */
/*  the adjacency list for testing purposes.               */
/***********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "part.h"
#include "read_graph.h"
#include <math.h>

#define MAX_ADJ(n)  	( (n) * ((n)-1) )
#define MAX_LINE	1000 

long random();
char *initstate(unsigned, char *, int);
void create_geom_graph(double);
int create_reg_graph(int, int, int, int);
int create_unbalanced_reg_graph(int, int);
void read_peinado(char *);

typedef struct plist *PLISTPTR;

typedef struct plist {
        struct adj_node *n;
        PLISTPTR next;
} PLIST;

static PLIST p_dummy;
static PLISTPTR p_head = &p_dummy;

static long state1[32] = {
               3,
               0x9a319039, 0x32d9c024, 0x9b663182, 0x5da1f342,
               0x7449e56b, 0xbeb1dbb0, 0xab5c5918, 0x946554fd,
               0x8c2e680f, 0xeb3d799f, 0xb11ee0b7, 0x2d436b86,
               0xda672e2a, 0x1588ca88, 0xe369735d, 0x904f35f7,
               0xd7158fd6, 0x6fa6f051, 0x616e6b96, 0xac94efdc,
               0xde3b81e0, 0xdf0a6fb5, 0xf103bc02, 0x48f340fb,
               0x36413f93, 0xc622c298, 0xf5a42ab8, 0x8a88d77b,
               0xf5ad9d0e, 0x8999220b, 0x27fb47b9
               };


/*static long state1[32] = {
               3,
               0x4a319039, 0x32d9c024, 0x4b663182, 0x5da1f342,
               0x7449e56b, 0x4eb1dbb0, 0x4b5c5918, 0x446554fd,
               0x4c2e680f, 0x4b3d799f, 0x411ee0b7, 0x2d436b86,
               0x4a672e2a, 0x1588ca88, 0x4369735d, 0x404f35f7,
               0x47158fd6, 0x6fa6f051, 0x616e6b96, 0x4c94efdc,
               0x4e3b81e0, 0x4f0a6fb5, 0x4103bc02, 0x48f340fb,
               0x36413f93, 0x4622c298, 0x45a42ab8, 0x4a88d77b,
               0x45ad9d0e, 0x4999220b, 0x27fb47b9
               };
*/

struct adj_node *adj_list[MAX_VERTICES] ;
int deg[MAX_VERTICES];     /** number of edges for each vertex   **/

extern double drand48();
extern void   srand48(long);

extern int num;
extern int Regular_Graphs; 	/* flag - create these or not */
extern int Unbalanced_Regular_Graphs;
extern int Unbalanced_Random_Graphs;
extern double prob;
extern long seed;
extern int numedges;

extern FILE *outfile, *efile ;

short **adj_mat;

static int Counter;
static struct adj_node *pool;

int Done_With_Initial_Graph;

void insert_plist(struct adj_node *n, PLISTPTR p_list)
{
	PLISTPTR ptemp = (PLISTPTR) malloc(sizeof(PLIST));
	
	ptemp->n = n;
	ptemp->next = p_list->next;
	p_list->next = ptemp;
}

void free_plist(PLISTPTR p_list)
{
	PLISTPTR temp, temp1;
	
	if (!p_list) return;
	temp1 = p_list->next;
	while (temp1 != NULL) {
		temp = temp1;
		temp1 = temp1->next;
		free((char *) temp->n);
		free((char *) temp);
	}
}


/*****************************************************/
/*******  initialize random number generator  ********/
/*****************************************************/
void randinit(unsigned cell)
{
	initstate(cell,(char *)state1,128);  /* 4/27/93                  */
	/*srand48(cell);*/                   /* uncommented 1/28/94      */
					     /* backtrack has been using */
					     /* unseeded lrand48() until */
					     /* now.			 */
}

double rrandom()
{
	return(random() / (double) MAX_INT);    /* 4/27/93 */
	/*return drand48();*/
}

double rrandom2()
{
	return(random() / ((double) MAX_INT/2.0));
	/*return drand48();*/
}

int irandom(int modulus)
{
	return (random()%modulus);
}

/*******************************************************/
/** adj(x, y)					      **/
/**		return TRUE if x is adjacent to y     **/
/*******************************************************/
int adj(int x, int y)
{
	struct adj_node *ptr;
	
	for (ptr=adj_list[x]; ptr; ptr= ptr->next)
		if (ptr->node_no == y)
			return(TRUE);
	return(FALSE);
}

void move_to_front(int node, int neighbor)
{
	struct adj_node *ptr, *temp;
	
	for (ptr=adj_list[node]; ptr && ptr->next; ptr= ptr->next)
		if (ptr->next->node_no == neighbor) {
			temp = ptr->next;
			ptr->next = temp->next;
			temp->next = adj_list[node];
			adj_list[node] = temp;
			return;
		}
}

/*************************************************************************/
/** INITIALIZE_ADJ_MATRIX  						**/
/*************************************************************************/
/** 	used only when creating regular graphs				**/
/**	FROM SetPlayer							**/
/*************************************************************************/
void init_adj_mat (short ***adj_mat)
{
	extern int num;
	int i, j;


	*adj_mat = (short **) malloc(sizeof(short *) * num);
	for ( i = 0; i < num; i++ ) {
		(*adj_mat)[i] = (short *) malloc(sizeof(short) * num);
	}
	for ( i = 0; i < (num - 1); i++ ) {
		(*adj_mat)[i][i] = 0;
		for ( j = (i + 1); j < num; j++ )
			(*adj_mat)[i][j] = (*adj_mat)[j][i] = 0;
	}
	(*adj_mat)[(num - 1)][(num - 1)] = 0;
	/*perror("init_adj_mat:");*/
}

void free_adj_mat (short **adj_mat, int reset)
{
	extern int deg[];
	extern int num;
	int i;

	if (reset) {
		for ( i = 0; i < num; i++ ) 
			deg[i] = 0;
	}

	for ( i = 0; i < num; i++ ) {
		free((char *) adj_mat[i]);
		printf(",");
	} 
	printf("\n");
	/*perror("free_adj_mat: before free((char *) adj_mat)");*/
	free((char *) adj_mat);
	/*perror("free_adj_mat: after free((char *) adj_mat)");*/
}

int expected_sum_degrees(double factor)
{
	extern int num;
	extern int Geometric_Graphs;
	extern int Regular_Graphs; 
	extern int Unbalanced_Regular_Graphs;
	extern int Unbalanced_Random_Graphs;
	extern int Regular_Graphs;
	extern int Leaves;
	extern int Reg_Degree;
	extern int Reg_Degree2;
	extern double prob, prob2;

	double result;
	double offset = prob * factor;
	
	if (Geometric_Graphs) {
		result = (prob + offset) * (prob + offset) * num * num * M_PI;
	} else if (Regular_Graphs) {
		result = num * Reg_Degree;
	} else if (Unbalanced_Regular_Graphs) {
		result = 2 * (num/2 + 1) * Reg_Degree + num * Reg_Degree2;
	} else if (Unbalanced_Random_Graphs) {
		result = 2 * ((prob+offset) * (num/2 +1) * (num/2)) +
			 num * (prob2+offset) * (num/2 +1);
	} else /* random graphs */
		result = (prob+offset) * num * (num-1);

	result = result + (2 * Leaves * num);
	return (int) result;
}
		
/*****************************************************************************/
/***** old allocation ********************************************************/
/*****************************************************************************/
/**	if (Leaves)
     		pool = (struct adj_node *)calloc(MAX_ADJ(num) + 2*(Leaves*num),
						 sizeof(struct adj_node));
	else
     		pool = (struct adj_node *)calloc(MAX_ADJ(num),
						 sizeof(struct adj_node));
**/
/*****************************************************************************/

/*******************************************************/
/**************  add_adj_entry() ***********************/
/*******************************************************/
void add_adj_entry(int to, int vertex)
{  
   extern int Regular_Graphs;
   extern int Done_With_Initial_Graph;

   double offset_factor = 0.01;
   static int size_allocd;
   struct adj_node *p1;

   if (Counter == 0) {
	/*perror("before adj alloc");*/
	size_allocd = expected_sum_degrees(offset_factor);
	while ((pool = (struct adj_node *)
			calloc(size_allocd, sizeof(struct adj_node))) == NULL) {
		offset_factor -= 0.05;
		size_allocd = expected_sum_degrees(offset_factor);
	}
	/*perror("after adj alloc");*/
	if (Regular_Graphs || Unbalanced_Regular_Graphs) {
		init_adj_mat(&adj_mat);
	}
	p_head->next = NULL;
   }

   if (Counter < size_allocd) {
   	p1 = &pool[Counter++];
   } else {
	p1 = (struct adj_node *)calloc(1,sizeof(struct adj_node));
	insert_plist(p1, p_head);
   }
   p1->node_no = vertex;
   p1->next = adj_list[to];
   adj_list[to] = p1;
   deg[vertex]++;
   if ((Regular_Graphs||Unbalanced_Regular_Graphs)&&!Done_With_Initial_Graph)  
		adj_mat[vertex][to] = 1;
}

/**************************************************************/
/** ADD_LEAVES()  *********************************************/
/**************************************************************/
/** 		For any type of graph, this routine adds     **/
/**		<Leaves> new leaves to each vertex. <Leaves> **/
/**		is a command-line option.		     **/
/**************************************************************/
void add_leaves()
{
	extern int Leaves;
	extern int Done_With_Initial_Graph;
	int i,j,k;
	
	Done_With_Initial_Graph = TRUE;
	for (i=0; i<num; i++)
		for (j=0; j<Leaves; j++) {
			k = (j+1)*num + i;
			add_adj_entry(i, k);
			add_adj_entry(k, i);
			numedges++;
		}
	num = num + (num * Leaves);
}
	
void output_leda(FILE *fp)
{
	int i, n;
	struct adj_node *point;

	fprintf(fp, "LEDA.GRAPH\n");
	fprintf(fp, "vertex\n");
	fprintf(fp, "int\n");
	fprintf(fp, "%d\n", num);
	for (i=1; i<=num; i++)
		fprintf(fp, "%d 0 0 0 0 0\n", i);
	fprintf(fp, "%d\n", numedges);
	for(i=0;i<num;i++) {
		for (point = adj_list[i]; point; point = point->next) {
			n = point->node_no;
			if (i < n)
				fprintf(fp,"%d %d 1\n", i+1, n+1);
		}
	}
}

int create_unbalanced_random_graph(double p1, double p2)
{
	int i, j;
	double p;

	for(i=0;i<(num/2);i++) {
		for(j=i+1;j<(num/2);j++) {
			if((p=rrandom()) < p1) {
				add_adj_entry(i,j);
				add_adj_entry(j,i);
				numedges++;
			}
		}
	}
	for(i=(num/2);i<num;i++) {
		for(j=i+1;j<num;j++) {
			if((p=rrandom()) < p1) {
				add_adj_entry(i,j);
				add_adj_entry(j,i);
				numedges++;
			}
		}
	}
	for(i=0;i<(num/2);i++) {
		for(j=(num/2);j<num;j++) {
			if((p=rrandom()) < p2) {
				add_adj_entry(i,j);
				add_adj_entry(j,i);
				numedges++;
			}
		}
	}
	return(TRUE);
}

int get_vertex(FILE *fp, int v_num)
{
        int i, next_int, int_read=FALSE;
        char str[MAX_LINE];
        char *p;
        char *str_int;

        p=fgets(str, MAX_LINE, fp);
        str_int=strtok(str," ");        /* degree stored explicitly - we omit */
        if (p) {
                str_int=strtok(NULL," ");
                while (str_int && (str_int[0] != '\n')) {
                        if ((i=sscanf(str_int,"%d", &next_int)) == 0)
                                printf(" file error: %s not integer\n",str_int),
                                exit(1);
                        add_adj_entry(v_num, next_int);
                        int_read = TRUE;
                        str_int=strtok(NULL," ");
                }
        }
        if (int_read) return(TRUE);
        return(FALSE);
}

/*******************************************************/
/***************  create_graph()  **********************/
/*******************************************************/
long create_graph()
{
	extern int Output_Leda;
	extern FILE *lfile;
	extern int Geometric_Graphs;
	extern int Peinado_Graphs;
	extern int Leaves;
	extern int Reg_Degree;
	extern int Reg_Degree2;
	extern double prob;
	extern double prob2;
	extern int Done_With_Initial_Graph;

	int i,j;
	double p;

	Done_With_Initial_Graph = FALSE;
	Counter = 0;
	numedges = 0 ;
	randinit(seed) ;
	for(i=0;i<(num+(Leaves*num)); i++) {
		adj_list[i] = NULL ;
		deg[i] = 0 ;
	}
	if (Geometric_Graphs) {
		create_geom_graph(prob);
		if (Leaves)  add_leaves();
		return(numedges);
	}
	if (Regular_Graphs) {
		if (create_reg_graph(Reg_Degree, 0, num-1, 0)) {
			if (Leaves)  add_leaves();
			return(numedges);
		} else {
			printf("Error: unable to create regular graph");
			exit(489);
		}
	}
	if (Unbalanced_Regular_Graphs) {
		if (create_unbalanced_reg_graph(Reg_Degree, Reg_Degree2)) {
			if (Leaves)  add_leaves();
			return(numedges);
		} else {
			printf("Error: unable to create unbal. regular graph");
			exit(490);
		}
	}
	if (Unbalanced_Random_Graphs) {
		if (create_unbalanced_random_graph(prob, prob2)) {
			if (Leaves)  add_leaves();
			return(numedges);
		} else {
			printf("Error: unable to create unbal. rand graph");
			exit(491);
		}
	}

	if(efile) { /** read edges from external file **/
		int deg_sum = 0;
		char *p;
		char str[MAX_LINE];

		rewind(efile) ; /** positition at beginning of file **/
        	p=fgets(str, MAX_LINE, efile);	/*get rid of first line*/
        	for (i=0; i<num; i++) {
			if (!get_vertex(efile, i))
				fprintf(outfile, "Error in reading Rao file\n");
		}
		for (i=0; i<num; i++)
			deg_sum += deg[i];
		numedges = deg_sum / 2;
	} else if (Peinado_Graphs) { 
		extern F_EDGE edge_list[];
		extern int nr_dir_edges;
		int i;

		for (i=0; i<nr_dir_edges; i++)
			add_adj_entry(edge_list[i].v1, edge_list[i].v2);
		numedges = nr_dir_edges/2;
	} else { /** generate random graph of given edge probability **/
		for(i=0;i<(num-1);i++) {
			for(j=i+1;j<num;j++) {
				if((p=rrandom()) < prob) {
					add_adj_entry(i,j);
					add_adj_entry(j,i);
					numedges++;
				}
			}
		}
	}
	add_leaves();

	if (Output_Leda) {
		output_leda(lfile);
		exit(0);
	}
	return(numedges);
}

/***************************************************/
/******   free_adj_list() **************************/
/***************************************************/
void free_adj_list() 
{
	extern int numedges;
	int i;

	free((char *) pool);
	free_plist(p_head);
	Done_With_Initial_Graph = FALSE;
	Counter = 0;
	numedges = 0 ;
	for(i=0;i<num;i++) 
		adj_list[i] = NULL;
}

     
/*******************************************************/
/***************  print adjacency list  ****************/
/*******************************************************/
/*---      routine used for testing purposes        ---*/
void print_adj_list()
{  struct adj_node *point;
   int i;
   fprintf(outfile,"adj list:\n");
   for(i=0;i<num;i++) {
      point = adj_list[i];
      fprintf(outfile,":%2d:",i);
      while(point != NULL) {
	 fprintf(outfile,"%d ",point->node_no);
	      point = point->next;
       }
      fprintf(outfile,"\n");
    }
}

int find_max_degree()
{
	int i, maxdeg = 0;

	for (i=0; i<num; i++)
		if (deg[i] > maxdeg)
			maxdeg = deg[i];
	return maxdeg;
}
