/**************************************************************/
/*                     Graph Paritioning                      */
/*                            gr.c                            */
/*                                                            */
/*                   Author:  James P. Fox                    */
/*                   Advisor: Mark Goldberg                   */
/*                                                            */
/*                       Masters Project                      */
/*               Rensselaer Polytechnic Institute             */
/*                     Summer, Fall, 1990                     */
/**************************************************************/
/*     This program impliments a "near-greedy" approach to    */
/*  graph partitioning.   It was designed in conjuction with  */
/*  the implementaion of the Kernighan & Lin algorithm and    */
/*  set up so as to be easily tuned using the K-L results.    */
/*  The routine evaluate_non_greedy in the module evalng.c    */
/*  performs the analysis and determines when to allow        */
/*  non-greedy steps in this process.  The program accepts    */
/*  command options to perfrom the maximum or minimum number  */
/*  of cuts and can take input from a file or from standard   */
/*  input.  Output can be sent to the screen or optionally to */
/*  a disk file.  A command option is also provided to show a */
/*  graphed result of the greedy/non-greedy steps used to     */
/*  to create the partitions.                                 */
/*                                                            */
/*      The following input is required:                      */
/*         number of vertices             1..1000             */
/*         edge probability               0.0001..0.9999      */
/*         random number seed             0..2147483647       */
/*         post-processing sort type      0..8                */
/*         look-ahead                     0 - no              */
/*                                        1 - yes             */
/*         confirm process of this data   y/n                 */
/*                                                            */
/*      The following additionial input is required for       */
/*         different reordering methods:                      */
/*      5,6,8:                                                */
/*         expected ind. set size         1..1000             */
/*         limit of the bounds            0..999999           */
/*         number of bounds to enter      1..1000             */
/*         each individual bounds         1..1000             */
/*      7,8:                                                  */
/*         tie situation:                 1 - highest degree  */
/*                                        2 - lowest degree   */
/*                                                            */
/**************************************************************/

#include <stdio.h>
#include <ctype.h>
#include "part.h"


/*****************  input data  ***********************/
extern int num ;              /** number of vertices **/
extern float prob;            /** edge probability   **/
extern long seed ;            /** seed **/
extern int num_passes;        /** number of times to apply random greedy algorythm. **/
extern int sort_type;         /** processing sort type **/
extern int look_ahead;        /** perform look-ahead when computing costs **/
extern int concurrent_even;   /** concurrent even selection **/
extern int lmt ;              /** independent set limit **/
extern int expected;          /** independent set expected **/
extern int bounds[MAX_IND_SIZE]; /** independent set bounds **/
extern int num_to_avg;        /** number of vertices to average non-greedy **/
extern int max_nongreedy ;    /** maximum number of non-greedy steps allowed **/
/*******************************************************/

extern struct adj_node *adj_list[MAX_VERTICES] ;
extern int deg[MAX_VERTICES];     /** number of edges for each vertex   **/
extern int num_vertices[], hold_num_vertices[] ;
extern int greedy, nongreedy ;
extern int post_tots[];

int vertices[MAX_VERTICES] ;
int vertex_loc[MAX_VERTICES];       /** vertex location                   **/
int vertex_processed[MAX_VERTICES]; /** unprocessed vertices              **/
int node_cost[2][MAX_VERTICES];     /** cost of vertex in each partition  **/
int keep_vertex[MAX_VERTICES];      /** last best partitioning            **/
int keep_vertices[MAX_VERTICES];    /** last best ordering                **/
int currentcost, oldcost, keepno;   /** maintain current cost information **/
int output_short, output_detail, output_adjacent;  /** output flags  **/

float min_mult ;
extern int read_file_mode;
int p_graph, mode;
int alg_type ;
char efilename[20];
FILE *infile, *efile ;
extern FILE *outfile, *tfile, *pfile ;

long numedges;
float time, t, tottime ;
int result, iter, tbestcost ;
extern int num_degree_nongreedy ;
extern int end_of_file,process ;
extern int max_size;
extern int tie1,tie2;
extern int hold_tie_1[], hold_tie_2[] ;
/*********************************************************/
/********  compute the cost of the current graph  ********/
/*********************************************************/
int compute_total_cost()
{  struct adj_node *point;
   int ccost, i, tloc;
   ccost = 0 ;
   for(i=0;i<num;i++) {  /*--- examine each vertex ---*/
      /*--- compute the cost for each node to be in each partition ---*/
      node_cost[SET_A][i] = node_cost[SET_B][i] = 0;
      tloc = vertex_loc[i];
      point = adj_list[i];
      while(point != NULL) {
         if(vertex_loc[point->node_no] == tloc) 
            node_cost[!tloc][i]++;
         else {
            node_cost[tloc][i]++;
            if(tloc == SET_A) /*** only count one vertex of the edge ***/
               ccost++;
	  }
         point = point->next;
       }
    }
   /*--- if minimize mode include cost for difference in partition size ---*/
   if(mode == MODE_MIN)
      ccost = ccost + min_mult *
                 ABS(num_vertices[SET_A] - num_vertices[SET_B]);
   return ccost;
 }

/******************************************/
/***********  GET VERTEX COSTS  ***********/
/******************************************/
int get_vertex_costs(n,costa,costb)
int n, *costa, *costb;
{  
   struct adj_node *point;
   *costa = *costb = 0 ;
   point = adj_list[n];
   while(point != NULL) {
      if(vertex_processed[point->node_no])
         if(vertex_loc[point->node_no] == SET_A)
            (*costb)++;
         else
            (*costa)++;
      point = point->next ;
    }
   return(*costa - *costb) ;
 }

/******************************************/
/**************  PLACE VERTEX  ************/
/******************************************/
void gr_place_vertex(n,set)
int n,set;
{  struct adj_node *p;
   vertex_loc[vertices[n]] = set;
   currentcost += node_cost[set][vertices[n]];
   vertex_processed[vertices[n]] = TRUE;
   /**  update node "costs" of all adjacent vertices **/
   p =adj_list[vertices[n]];
   while(p != NULL) {
      node_cost[!set][p->node_no]++;
      p = p->next ;
    }
   num_vertices[set]++;
 }

/******************************************/
/*************  REMOVE VERTEX  ************/
/******************************************/
void gr_remove_vertex(n)
int n;
{  struct adj_node *p;
   int set ;
   set = vertex_loc[vertices[n]];
   vertex_processed[vertices[n]] = FALSE;
   currentcost -= node_cost[set][vertices[n]];
   /**  update node "costs" of all un-processed adjacent vertices **/
   p = adj_list[vertices[n]];
   while(p != NULL) {
      node_cost[!set][p->node_no]--;
      p = p->next ;
    }
   num_vertices[set]--;
 }

/********************************************************/
/**********  compute future costs for look-ahead  *******/
/********************************************************/
void compute_future(v,futa,futb)
int v, *futa, *futb;
{  struct adj_node *p;
   register i;
   *futa = *futb = 0 ;
   p = adj_list[v];
   while(p!=NULL) {
     i = p->node_no;
     if(!vertex_processed[i])
        if(node_cost[SET_A][i] > node_cost[SET_B][i])
           (*futb)++;
        else
           if(node_cost[SET_A][i] < node_cost[SET_B][i])
              (*futa)++;
     p = p->next ;
   }
 }


/********************************************************/
/************  perform near-greedy algorithm  ***********/
/********************************************************/
void near_greedy_pass()
{  register i, n,j, k;
   int futa, futb;
   int done,exit,bestcost;
   int num_nongreedy ;
   int lastmove[MAX_VERTICES];
   int diff[MAX_VERTICES], localsum,d ;
   int num_deg_ng[MAX_DEGREE_NONGREEDY+1];
   float v ;
   struct adj_node *p;
   int hcost, changed, a, b, set ;
   for(i=0;i<num;i++) {
      vertex_processed[i] = FALSE;
      node_cost[SET_A][i] = node_cost[SET_B][i] = 0 ;
    }
   init_eval_non_greedy();   
   for(i=0;i<=num_degree_nongreedy;i++)
      num_deg_ng[i] = 0 ;
   lastmove[0] = 0;
   currentcost = 0 ;
   num_nongreedy = 0 ;
   futa = futb = 0 ;
   num_vertices[SET_A] = num_vertices[SET_B] = 0 ;
   t = log_time();
   n = 0;
   bestcost = VERY_LOW ;
   while(n >= 0) {
      if(n < num) { /** move down a level **/
         if(sort_type == 7) /** concurrent reordering **/
            vertices[n] = get_next_concurrent();
         if(sort_type == 8 && n < num-max_size)
            vertices[n] = get_next_concurrent();
         hold_tie_1[n] = tie1 ;
         hold_tie_2[n] = tie2 ;
         j = vertices[n]; /** place greedy **/
          if(look_ahead) {
             compute_future(j,&futa,&futb);
             diff[n] = ABS(node_cost[SET_A][j] + futa 
                      -node_cost[SET_B][j] - futb);
          }
          else
             diff[n] = ABS(node_cost[SET_A][j] - node_cost[SET_B][j]);
         if(diff[n] > num_degree_nongreedy)
            diff[n] = num_degree_nongreedy;
         if(mode == MODE_MAX)  /** maximize **/
            if(node_cost[SET_A][j] + futa >= node_cost[SET_B][j] + futb)
               gr_place_vertex(n,SET_A);
            else
               gr_place_vertex(n,SET_B);
         else {  /** minimize **/
            k = (int) ((float)(num_vertices[SET_A] - num_vertices[SET_B]) *
                 min_mult);
            if(node_cost[SET_A][j] + futa + k > node_cost[SET_B][j] + futb)
               gr_place_vertex(n,SET_B);
            else
               gr_place_vertex(n,SET_A);
	  }
         lastmove[n] = FALSE;
         n++;
       }
      else { /** backtrack **/
         /** check for best "cuts" so far **/
         if(n == num && currentcost > bestcost) {
	    for(i=0;i<num;i++) {
	       keep_vertex[i] = vertex_loc[i] ;
               keep_vertices[i] = vertices[i];
	     }
	    bestcost = currentcost ;
	  }
         n--; 
         exit = FALSE ;
         while(!exit) {
            if(n < 0)
               exit = TRUE;
            else {
               gr_remove_vertex(n);
               if(!lastmove[n] && 
                    evaluate_non_greedy(n,num_deg_ng[diff[n]])) {
                  /** take non-greedy step **/
                  num_deg_ng[diff[n]]++;
                  exit = TRUE;
                  gr_place_vertex(n,!vertex_loc[vertices[n]]);
                  lastmove[n] = TRUE;
                  n++;
		}
               else {
                  if(lastmove[n])  /** undo non-greedy count **/
                     num_deg_ng[diff[n]]--;
                  n--;
		}
	     }
	  }
       }
    }
   t = log_time();
   currentcost = bestcost ;
   for(i=0;i<num;i++) {
      vertex_loc[i] = keep_vertex[i];
      vertices[i] = keep_vertices[i];
    }
   if(currentcost != compute_total_cost()) {
      printf("####internal program error #2001\n");
      printf("     currentcost = %d\n",currentcost);
      printf("     actual cost = %d\n",compute_total_cost());
    }
   post_processing(TRUE);
 }

/*******************************************************/
/****************  print out cost info  ****************/
/*******************************************************/
/*---      routine used for testing purposes        ---*/
void print_costs()
{  int i;
   fprintf(outfile,"       loc:");
   for(i=0;i<num;i++)
      fprintf(outfile,"%3d",vertex_loc[i]);
   fprintf(outfile,"\n");
   fprintf(outfile,"Cost SET_A:");
   for(i=0;i<num;i++)
      fprintf(outfile,"%3d",node_cost[SET_A][i]);
   fprintf(outfile,"\n");
   fprintf(outfile,"Cost SET_B:");
   for(i=0;i<num;i++)
      fprintf(outfile,"%3d",node_cost[SET_B][i]);
   fprintf(outfile,"\n");
 }

/***********************************************/
/***  print the text graphs of computed data ***/
/***********************************************/
void print_graphs()
{  int i, aplot[MAX_VERTICES];
   fprintf(outfile,"\nDegree of Vertices\n");
   print_text_graph(post_tots,num,"Greedy","Non-Greedy",outfile,FALSE);
   fprintf(outfile,"\nGreedy/Non-Greedy Vertices\n");
   for(i=0;i<num;i++) {
      if(post_tots[i] > 0)
	 aplot[i] = 1;
      else if(post_tots[i] < 0)
	 aplot[i] = -1;
      else
	 aplot[i] = 0 ;
    }
   print_text_graph(aplot,num,"Greedy","Non-Greedy",outfile,FALSE);
 }


/******************************************************/
/***********  print current vertex locations  *********/
/******************************************************/
/*---        this routine is used for testing      ---*/
void print_locs()
{  int i;
   printf("locs:");
   for(i=0;i<num;i++)
      printf("%d ",vertex_loc[i]);
   printf("\n");
 }

/**********************************************************/
/**********************************************************/

main(argc,argv) 
   int argc ;
   char *argv[];
{  int i;
   alg_type = ALG_TYPE_NEAR_GREEDY ;
   clear_summary();
   process_parameters(argc,argv);
   end_of_file = FALSE;
   while(!end_of_file) {
      if(!read_data()) {
         printf("invalid data input, program terminated....\n");
         exit(1);
       }
      if(process) {
         create_graph() ;
         process_reorder() ;
         if(num<=25)
            print_adj_list();
         tottime = 0.0 ;
         fprintf(outfile,"OUTPUT:\n");
         fprintf(outfile,"   edges:%6d\n",numedges);
         if(mode == MODE_MAX)
            fprintf(outfile,"Maximize number of edges cut\n");
	 else
	    fprintf(outfile,"Minimize number of edges cut\n");
         if(!output_short) {
            fprintf(outfile,"-------------------------------------------------\n");
            fprintf(outfile,"PASS  TIME   CUTS   GREEDY   NON-GR.  SET_A SET_B\n");
            fprintf(outfile,"-------------------------------------------------\n");
	  }
         near_greedy_pass();
         if(num <= 25) /** test mode **/
            print_locs();
         add_to_summary();
         if(output_short)
	    fprintf(outfile,"Cuts in best partition: %5d\n",currentcost);
	 else
            fprintf(outfile,"%4d %6.2f  %5d   %5d %5d %3d%%  %4d  %4d\n",
               i+1,t,currentcost,greedy,nongreedy,
               (nongreedy*100)/(greedy+nongreedy),
               num_vertices[SET_A],num_vertices[SET_B]);
         print_non_greedy();
         if(p_graph)
            print_graphs();
         fflush(outfile);
         free_adj_list() ;
       }
    }
   num = 0; prob = 0.0; /** force final summary **/
   print_summary();
}
