/**************************************************************/
/*                     Graph Paritioning                      */
/*                            kl.c                            */
/*                                                            */
/*                   Author:  James P. Fox                    */
/*                   Advisor: Mark Goldberg                   */
/*                                                            */
/*                       Masters Project                      */
/*               Rensselaer Polytechnic Institute             */
/*                     Summer, Fall, 1990                     */
/**************************************************************/
/*     This program impliments the Kernighan & Lin algorithm  */
/*  partitioning. (The Bell System Technical Journal, Feb.,   */
/*  1970, pp. 291-307.)  This algorithm attempts to find the  */
/*  partitions which provide the maximum/minimum number of    */
/*  edges between the partitions.  The program accepts        */
/*  command options to perfrom the maximum or minimum number  */
/*  of cuts and can take input from a file or from standard   */
/*  input.  Output can be sent to the screen or optionally to */
/*  a disk file.  A command option is also provided to show a */
/*  graphed result of the greedy/non-greedy steps used to     */
/*  to create the partitions.                                 */
/*                                                            */
/*      The following input is required:                      */
/*         number of vertices             1..1000             */
/*         edge probability               0.0001..0.9999      */
/*         random number seed             0..2147483647       */
/*         number of applications of kl   1..99999            */
/*         post-processing sort type      0..8                */
/*         look-ahead                     0 - no              */
/*                                        1 - yes             */
/*         confirm process of this data   y/n                 */
/*                                                            */
/*      The following additionial input is required for       */
/*         different reordering methods:                      */
/*      5,6,8:                                                */
/*         expected ind. set size         1..1000             */
/*         limit of the bounds            0..999999           */
/*         number of bounds to enter      1..1000             */
/*         each individual bounds         1..1000             */
/*      7,8:                                                  */
/*         tie situation:                 1 - highest degree  */
/*                                        2 - lowest degree   */
/*                                                            */
/*      After input is received the graph is created and      */
/*  processed using the k-l algorithm.  After each            */
/*  application the best (min/max) partition information is   */
/*  printed including information on greedy/non-greedy steps. */
/*  After the last application of k-l the best partition is   */
/*  printed, along with an optional graph of greedy/non-      */
/*  greedy steps.                                             */
/**************************************************************/

#include <stdio.h>
#include <ctype.h>
#include "part.h"

int output_rao();
int one_walk(int, int, double *);
void init_walk(int);
int find_max_degree();
void output_gw();
void output_chaco();
void output_gp();
void output_ng();
void dump_cutfiles(int *, int, int);
void line_randomize();
void open_greedy_outfile(int pass);
void close_greedy_outfile();
void save_and_postproc(int, int, int, int *, int *);
void post_processing(int, int, int);
void randinit(unsigned);
void print_text_graph(int *, int, char *, char *, FILE *, int);
void prompt(char *);
void find_augmenting_paths();
void kl_pass(int);



/*****************  input data  ******************************************/
extern int num ;                /** number of vertices 			**/
extern int num_passes;		/** if not Heuristic_Only then count    **/
extern double cpu_seconds;       /** time to run the k-l algorithm.      **/
/*************************************************************************/

extern int greedy, nongreedy ;
extern int hold_post_tots[];
extern int hold_greedy, hold_nongreedy;
extern int num_vertices[], hold_num_vertices[] ;
extern int read_file_mode;
extern int end_of_file,process ;
extern int Output_Gnuplot;

extern FILE *outfile;
extern struct adj_node *adj_list[MAX_VERTICES] ;
extern LISTPTR nodes;

int Done_With_Alg;		    /** TRUE after kl_main_loop()     **/
				    /** or sim_annealing()            **/
				    /** used in postproc.c            **/
int currentcost, keepno;            /** maintain current cost info    **/
int currentminsize;   		    /** size of smaller partition     **/
int tbestcost;
int tbestminsize;
int output_short, output_detail, output_adjacent;  /** output flags   **/
int result; 

int iter;				/** #iterations in one pass(run) **/
int Num_Runs;				/** #runs of entire alg	 	 **/

int p_graph, mode;
int node_cost[2][MAX_VERTICES];     	/** cost of each v. in each part, **/
					/** measured by the number of adj **/
					/** to vertices already placed.   **/
int hold_ng_cost[2][MAX_VERTICES];     	/** same cost, except only        **/
					/** nongreedy placed neighbors 	  **/
					/** are included.  Only recorded  **/
					/** for the best partitioning     **/
int vertices[MAX_VERTICES] ;
int vertex_loc[MAX_VERTICES];       	/** vertex location          	  **/
int vertex_processed[MAX_VERTICES]; 	/** unprocessed vertices     	  **/
int hold2_num_vertices[2];	     	/** numbers in best partition  	  **/
int hold_kl_loc[MAX_VERTICES];
int Path_Vertices[10];
int Starting_Locs[MAX_VERTICES]; /* for Simulated Annealing */
double Current_Temp;		 /* for Sim. Annealing also */

int numedges;
double min_mult;
double time_, tottime;
char efilename[20];
FILE *infile, *efile ;


/******************************************************************/
/************************  randomize vertices  ********************/
/******************************************************************/
void randomize_vertices() 
{
   extern LISTPTR nodes;
   int i, numa, numb;

   num_vertices[SET_A] = num_vertices[SET_B] = 0 ;
   numa = num/2 ;
   numb = num-numa;
   for(i=0;i<num;i++) {
      if(rrandom() < .5) {
            vertex_loc[i] = SET_A;
	    nodes[i].loc = SET_A;
	    Starting_Locs[i] = SET_A;
      } else {
            vertex_loc[i] = SET_B;
	    nodes[i].loc = SET_B;
	    Starting_Locs[i] = SET_B;
      }
      num_vertices[nodes[i].loc]++ ;
   }
}
/*
         if(numa > 0) { 
            numa--;
            vertex_loc[i] = SET_A;
	    nodes[i].loc = SET_A;
	    Starting_Locs[i] = SET_A;
	  }
         else {  
            numb--;
            vertex_loc[i] = SET_B;
	    nodes[i].loc = SET_B;
	    Starting_Locs[i] = SET_B;
	  }
       }
      else {
         if(numb > 0) {  
            numb--;
            vertex_loc[i] = SET_B;
	    nodes[i].loc = SET_B;
	    Starting_Locs[i] = SET_B;
	  }
         else {  
            numa--;
            vertex_loc[i] = SET_A;
	    nodes[i].loc = SET_A;
	    Starting_Locs[i] = SET_A;
	  }
       }
    }
}*/


/*********************************************************/
/********  compute the cost of the randomized  graph  ****/
/*********************************************************/
int compute_initial_cost(int node_cost[][MAX_VERTICES], int *vertex_loc)
{  struct adj_node *point;
   int ccost, i, tloc;
   ccost = 0 ;
   for(i=0;i<num;i++) {  /*--- examine each vertex ---*/
      /*--- compute the cost for each node to be in each partition ---*/
      node_cost[SET_A][i] = node_cost[SET_B][i] = 0;
      tloc = vertex_loc[i];
      point = adj_list[i];
      while(point != NULL) {
         if(vertex_loc[point->node_no] == tloc) 
            node_cost[!tloc][i]++;
         else {
            node_cost[tloc][i]++;
            if(tloc == SET_A) /*** only count one vertex of the edge ***/
               ccost++;
	  }
         point = point->next;
       }
    }
   return ccost;
}

/*********************************************************/
/********  compute the cost of the current graph  ********/
/*********************************************************/
int compute_costs()
{
	struct adj_node *point;
	int ccost = 0, i, loc;

	num_vertices[0] = num_vertices[1] = 0;
	for(i=0;i<num;i++) {
		node_cost[SET_A][i] = node_cost[SET_B][i] = 0;
	   	loc = nodes[i].loc;
		num_vertices[loc]++;
		for (point = adj_list[i]; point; point = point->next) 
			if(nodes[point->node_no].loc == loc) 
				node_cost[!loc][i]++;
			else {
				node_cost[loc][i]++;	
				if (loc == SET_A)
					ccost++;
			}
	}
	return(ccost);
}

int compute_cost(int *locs)
{
	struct adj_node *point;
	int ccost = 0, i;

    if (locs == NULL)
	for(i=0;i<num;i++) {
		if (nodes[i].loc == SET_A) {
			for (point = adj_list[i]; point; point = point->next)
				if(nodes[point->node_no].loc == SET_B)
					ccost++;
		}
	}
    else
	for(i=0;i<num;i++) {
		if (locs[i] == SET_A) {
			for (point = adj_list[i]; point; point = point->next)
				if(locs[point->node_no] == SET_B)
					ccost++;
		}
	}
    return(ccost);
}

int check_cost(int *locs, int currentcost)
{
	struct adj_node *point;
	int ccost = 0, i;

    if (locs == NULL)
	for(i=0;i<num;i++) {
		if (nodes[i].loc == SET_A) {
			for (point = adj_list[i]; point; point = point->next)
				if(nodes[point->node_no].loc == SET_B)
					ccost++;
		}
	}
    else
	for(i=0;i<num;i++) {
		if (locs[i] == SET_A) {
			for (point = adj_list[i]; point; point = point->next)
				if(locs[point->node_no] == SET_B)
					ccost++;
		}
	}
    if (ccost != currentcost) printf("error: act. cost: %d\n", ccost);
    return(ccost == currentcost);
}

int test1_total_cost(int *locs)
{
	struct adj_node *point;
	int ccost = 0, i;

	for(i=0;i<num;i++) {
		if (locs[i] == SET_A) {
			for (point = adj_list[i]; point; point = point->next)
				if(locs[point->node_no] == SET_B)
					ccost++;
		}
	}
	return(ccost);
}

/*********************************************************/
/********  compute the cost of the current graph  ********/
/*********************************************************/
int compute_total_cost(int *locs, int (*node_cost)[MAX_VERTICES], int *verts)
{  
   	int ccost, i; 

   	ccost = 0 ;
   	for(i=0;i<num;i++)  
         	if(locs[i] == SET_A) 
            		ccost += node_cost[SET_A][i];
         	else
            		ccost += node_cost[SET_B][i];
	ccost /= 2;

   	/*if(mode == MODE_MIN)
      		ccost = ccost + min_mult *
                ABS(verts[SET_A] - verts[SET_B]);*/
   	return ccost;
}
/*
int compute_total_cost(int node_cost[][MAX_VERTICES])
{  
   	int ccost, i; 

   	ccost = 0 ;
   	for(i=0;i<num;i++)  
         	if(vertex_loc[i] == SET_A) 
            		ccost += node_cost[SET_A][i];
         	else
            		ccost += node_cost[SET_B][i];
	ccost /= 2;

   	if(mode == MODE_MIN)
      		ccost = ccost + min_mult *
                ABS(num_vertices[SET_A] - num_vertices[SET_B]);
   	return ccost;
}
*/
/**********************************************************/
/********  move a vertex to other partition  **************/
/**********************************************************/
void move_vertex(int n)
{  register i;
   int from;
   struct adj_node *point ;
   from = vertex_loc[n];
   /*--- move vertex ---*/
   vertex_loc[n] = !from;
   /*--- compute updated costs for each vertex adjacent to it ---*/
   point = adj_list[n];
   while(point != NULL) {
      i = point->node_no;
      if(from) {
         node_cost[SET_A][i]--;
         node_cost[SET_B][i]++;
       }
      else {
         node_cost[SET_B][i]--;
         node_cost[SET_A][i]++;
       }
      point= point->next;
    }
   /*--- update number of vertices in each partition ---*/
   num_vertices[from]--;
   num_vertices[!from]++;
 }


/*************************************************/
/********  get best cost node to move  ***********/
/*************************************************/
int get_best_cost_node()
{  register i,j, n=0;
   int t, k, costdiff;
   if(mode == MODE_MAX) {
      costdiff = VERY_LOW ;
      for(i=0;i<num;i++) {
         if(!vertex_processed[i]) {
            t = vertex_loc[i] ;
            j = node_cost[!t][i] - node_cost[t][i];
            if(j > costdiff) {
               costdiff = j;
               n = i ;
	     }
	  }
       }
    }
   else {
      /*--- find minimum cost difference ---*/
      costdiff = VERY_HIGH;
      for(i=0;i<num;i++) {
         if(!vertex_processed[i]) {
            t = vertex_loc[i] ;
            k = ABS(num_vertices[!t] - num_vertices[t]) -
                ABS(num_vertices[!t] - num_vertices[t] + 2);
            j = node_cost[!t][i] - node_cost[t][i] - (min_mult*k);
            if(j< costdiff) {
               costdiff = j;
               n = i;
	     }
	  }
       }
    }
   currentcost = currentcost + costdiff;
   vertex_processed[n] = TRUE;
   return n;
}

int check_kl_path(int v, int path_index)
{      
        extern int Path_Vertices[];
        extern int Strict_Trees;
	int j;

        if (!adj(v, Path_Vertices[path_index]))
        return FALSE;

        if (!Strict_Trees)
                return TRUE;

        for (j=path_index-1; j>=0; j--) {
                if (adj(v, Path_Vertices[j]))
                        return FALSE;
        }
        return TRUE;
}

LISTPTR find_next_kl_path_vertex(int side, int path_length)
{
	extern int Current_Max_Bucket[2];
	extern int Current_Min_Bucket[2];
	extern LISTPTR *buckets[2];
	extern int Path_Vertices[]; 

	int path_index = path_length-1;
	int bnum = Current_Max_Bucket[side];
	LISTPTR p = buckets[side][bnum]->next;

	if (path_index < 0)  
		printf("error: path index < 0\n"), exit(24902350);

	while (1) {
		for (; p; p=p->next) {
		    if (check_kl_path(p->node_no, path_index)) {
				Path_Vertices[path_length] = p->node_no;
				return p;
		    }
		}
		if (--bnum < Current_Min_Bucket[side])
			break; 
		p = buckets[side][bnum]->next;
	}
	return NULL;
}

LISTPTR get_next_node()
{
	extern int Path_Size;
	extern LISTPTR bucket_extract_max(), bucket_extract_min();

	static int plength, side;
	LISTPTR node=NULL;

	if ((plength % Path_Size) == 0) {
		node = bucket_extract_max(LOCKED);
		side = node->loc;
		Path_Vertices[0] = node->node_no;
		plength = 1;
	} else {
		side = !side;
		if ((node=find_next_kl_path_vertex(side,plength))) {
			delete_from_bucket(node);
			node->lock = LOCKED;
			plength++;
		} else {
			node = bucket_extract_max(LOCKED);
			side = node->loc;
			Path_Vertices[0] = node->node_no;
			plength = 1;
		}
	}
	return node;
}

/**************************************************/
/*****  perform one pass of the k-l algorithm  ****/
/**************************************************/
#ifdef BUCKET

void kl_pass(int num_fixed)
{

	extern int alg_type;
	extern int View_Pass;
	extern int Path_KL;
	extern LISTPTR bucket_extract_max(), bucket_extract_min();

	register int i, loc;
	int pluses = 0;			/* # moves that improve cuts       */
	int minuses = 0;		/* # moves that worsen  cuts       */
	int neutrals = 0;		/* # moves that don't affect cuts  */
	int best_pluses = 0;		/* save value for best partition   */
	int best_minuses = 0;		/* save value for best partition   */
	int best_neutrals = 0;		/* save value for best partition   */
	int bestcost;
	int bestminsize = 0;
	int keep_vertex[MAX_VERTICES]; 
	char gain_record[MAX_VERTS];	/* record each KL move - pos or neg */
	LISTPTR node;
	
	currentcost = compute_costs();
	currentminsize = MIN(num_vertices[SET_A], num_vertices[SET_B]);
	reset_buckets(node_cost);

	if(mode == MODE_MAX) {
		bestcost = VERY_LOW ;
		for (i=0;i<num-num_fixed;i++) {			/* 3/20/93 */
			if (Path_KL)
				node = get_next_node();
			else 
				node = bucket_extract_max(LOCKED);
			if (View_Pass) {
				if (node->cost > 0) {
					pluses++;
					gain_record[i] = '+';
				} else if (node->cost == 0) {
					neutrals++;
					gain_record[i] = '_';
				} else	{
					minuses++;
					gain_record[i] = '-';
				}
			}
			keep_vertex[keepno++] = node->node_no;
			bucket_move_vertex(node);
			if(currentcost > bestcost) {
				bestcost = currentcost ;
				keepno =0;
				if (View_Pass) {
					best_pluses = pluses;
					best_minuses = minuses;
					best_neutrals = neutrals;
				}
			}
		}
	} else {
		bestcost = VERY_HIGH ;
		bestminsize = 1;
		for(i=0;i<num;i++) {
			node = bucket_extract_min(LOCKED);
			if (View_Pass) {
				if (node->cost > 0) {
					pluses++;
					gain_record[i] = '+';
				} else if (node->cost == 0) {
					neutrals++;
					gain_record[i] = '_';
				} else	{
					minuses++;
					gain_record[i] = '-';
				}
			}
			bucket_move_vertex(node);
			keep_vertex[keepno++] = node->node_no;
			if (mode == MIN_QUOTIENT) {
			  if( ((double) currentcost / (double) currentminsize)<=
			      ((double) bestcost / (double) bestminsize) ) {
					bestcost = currentcost ;
					bestminsize = currentminsize ;
					keepno =0;
					if (View_Pass) {
						best_pluses = pluses;
						best_minuses = minuses;
						best_neutrals = neutrals;
					}
			  }
			} else {
				if(currentcost < bestcost) {
					bestcost = currentcost ;
					keepno =0;
					if (View_Pass) {
						best_pluses = pluses;
						best_minuses = minuses;
						best_neutrals = neutrals;
					}
				}
			}
		}
	}

	for(i=keepno-1;i>=0;i--) {
		loc = nodes[keep_vertex[i]].loc;
		nodes[keep_vertex[i]].loc = !loc;
		num_vertices[loc]--;
		num_vertices[!loc]++;
	}
	if (View_Pass) {
		printf("KL ITERATION: ");
		printf(" Moves: %d, pos: %d, neutral: %d, negative: %d\n", 
					num-keepno, best_pluses,
					best_neutrals, best_minuses);
		for (i=0; i<num-keepno; i++)
			printf("%c", gain_record[i]);
		printf("/");
		for (i=num-keepno; i<num; i++)
			printf("%c", gain_record[i]);
		printf("\n");
	}
	currentcost = bestcost;
	currentminsize = bestminsize;
	if (Output_Gnuplot && alg_type == ALG_TYPE_KL)
		dump_cutfiles(vertices,0,num), exit(1);
}

#else

kl_pass()
{  register i, n;
   int bestcost ;
   int keep_vertex[MAX_VERTICES]; 

   for(i=0;i<num;i++)
      vertex_processed[i] = FALSE;
   if(mode == MODE_MAX) {  /*--- maximize mode ---*/
      bestcost = VERY_LOW ;
      for(i=0;i<num;i++) {
         keep_vertex[keepno++] = n = get_best_cost_node() ;
         move_vertex(n);
         if(currentcost >= bestcost) {
            bestcost = currentcost ;
            keepno =0;
	  }
       }
    }
   else {  /*---  minimize mode ---*/
      bestcost = VERY_HIGH ;
      for(i=0;i<num;i++) {
         keep_vertex[keepno++] = n = get_best_cost_node();
         move_vertex(n);
         if(currentcost <=bestcost) {
            bestcost = currentcost ;
            keepno = 0 ;
	  }
       }
    }

   /***  perform rollback of changes since last best value  ***/
   for(i=keepno-1;i>=0;i--)
      move_vertex(keep_vertex[i]);
   currentcost = bestcost;
  }

#endif
/*******************************************************/
/****************  print out cost info  ****************/
/*******************************************************/
/*---      routine used for testing purposes        ---*/
void print_costs()
{  int i;
   fprintf(outfile,"       loc:");
   for(i=0;i<num;i++)
      fprintf(outfile,"%3d",vertex_loc[i]);
   fprintf(outfile,"\n");
   fprintf(outfile,"Cost SET_A:");
   for(i=0;i<num;i++)
      fprintf(outfile,"%3d",node_cost[SET_A][i]);
   fprintf(outfile,"\n");
   fprintf(outfile,"Cost SET_B:");
   for(i=0;i<num;i++)
      fprintf(outfile,"%3d",node_cost[SET_B][i]);
   fprintf(outfile,"\n");
 }

/************************************************************/
/*****************   maximize edges cut   *******************/
/************************************************************/
void perform_maximize(int begin_fix, int end_fix, int *bestcost)  
{
	extern int Heuristic_Only;
	int num_fixed = end_fix - begin_fix;
	int *bcost = bestcost;
	int dummy;
	int oldcost = VERY_LOW;

	iter = 0;
	time_ = log_time();
	while(oldcost < currentcost) {
		oldcost = currentcost;
		iter++ ;
		kl_pass(num_fixed);
	}
	time_ = log_time();
	tottime = tottime + time_;

	/*****Check this - should we postproc after every up-down*******/

	if (!Heuristic_Only && !check_cost(NULL, currentcost)) 
		printf("*************************  program error #1 \n");
	save_and_postproc(mode, begin_fix, end_fix, bcost, &dummy);
}


/**************************************************/
/************   minimize edges cut   **************/
/**************************************************/
void perform_minimize(int begin_fix,int end_fix,int *bestcost,int *bestminsize)
{
	extern int Heuristic_Only;
	int num_fixed = end_fix - begin_fix;
	int oldcost = VERY_HIGH;
	int oldminsize = 1;

	iter = 0;
	time_ = log_time();
	currentcost = 0;
	currentminsize = 1;
	if (mode == MIN_QUOTIENT) {
		while (1) {
			iter++;
			kl_pass(num_fixed);
			if ( ((double) oldcost / (double) oldminsize) <=
			     ((double) currentcost / (double) currentminsize)) 
				break;
			oldcost = currentcost;
			oldminsize = currentminsize;
		}
	} else {
		while(oldcost > currentcost) {
			oldcost = currentcost;
			iter++;
			kl_pass(num_fixed);
		}
	}
	time_ = log_time();
	tottime = tottime + time_;
	if (!Heuristic_Only && !check_cost(NULL, currentcost)) 
		printf("*************************  program error #1 \n");
	save_and_postproc(mode, begin_fix, end_fix, bestcost, bestminsize);
}

void print_graphs()
{  int i, aplot[MAX_VERTS];
   fprintf(outfile,"\nDegree of Vertices\n");
   print_text_graph(hold_post_tots,num,"Greedy","Non-Greedy",outfile,FALSE);
   fprintf(outfile,"\nGreedy/Non-Greedy Vertices\n");
   for(i=0;i<num;i++) {
      if(hold_post_tots[i] > 0)
         aplot[i] = 1;
      else if(hold_post_tots[i] < 0)
         aplot[i] = -1;
      else
         aplot[i] = 0 ;
    }
   print_text_graph(aplot,num,"Greedy","Non-Greedy",outfile,FALSE);
   fprintf(outfile,"\nCumulative Degree of Vertices\n");
   print_text_graph(hold_post_tots,num,"Greedy","Non-Greedy",outfile,TRUE);
   fprintf(outfile,"\nCumulative Greedy/Non-Greedy Vertices\n");
   print_text_graph(aplot,num,"Greedy","Non-Greedy",outfile,TRUE);
}

/******************************************************/
/***********  print current vertex locations  *********/
/******************************************************/
/*---        this routine is used for testing      ---*/
void print_locs(int *locs)
{  int i;
   printf("locs:");
   for(i=0;i<num;i++)
      printf("%d ",locs[i]);
   printf("\n");
}

void print_node_locs()
{  
	extern LISTPTR nodes;
	int i;
	printf("locs:");
	for(i=0;i<num;i++)
		printf("%d ",nodes[i].loc);
	printf("\n");
}

void print_locs_to_file(int *locs)
/***********************************************************************/
/*** written to write the locations out to a file so that partitions ***/
/*** may be compared.  This is needed for the testing of placing     ***/
/*** one vertex differently, then comparing partitions (in bk)	     ***/
/***********************************************************************/
{  int i;
   char fname[80];
   FILE *lfp;

   prompt("Please enter locs filename");
   scanf("%s", fname);
   if ((lfp = fopen(fname, "w")) == NULL)
	printf("Error opening %s\n", fname), exit(45);
   fprintf(lfp, "%d\n", num);
   for(i=0;i<num;i++)
      fprintf(lfp, "%d ",locs[i]);
   fprintf(lfp, "\n");
}

void print_loc()
{  int i;
   printf("locs:");
   for(i=0;i<num;i++)
      printf("%d ",nodes[i].loc);
   printf("\n");
}

int compare_and_save(int pass, int *best_locs)
{
	int i, diff;
	static int last_locs[MAX_VERTS];

	if (pass == 0)
		for (i=0;i<num;i++)
			last_locs[i] = best_locs[i];
	diff = num;
	for (i=0; i<num; i++) {
		diff -= (best_locs[i] == last_locs[i]);
		last_locs[i] = best_locs[i];
	}
	if (num-diff < diff)
		diff = num-diff;
	return(diff);
}


void kl_init_loop(int alloc_b)
{ 
	extern int Output_Chaco;	/** flag - output for sandia's Chaco**/
	extern int Output_Rao;		/** flag - output for satish rao's **/
	extern int Output_GW;		/** flag - output goemans/william.**/
	extern int Output_GP;		/** flag - output graphpack or not**/
	extern int  Done_With_Alg;
        static char hyphens[] = "--------------------------------------------------------------------------------";

	Done_With_Alg = FALSE;
	if (alloc_b) {
		create_graph();
		if (Output_Chaco) 
			output_chaco(), exit(0);
		if (Output_Rao) 
			output_rao(), exit(0);
		if (Output_GW) 
			output_gw(), exit(0);		
		if (Output_GP) {
			output_gp();	
			output_ng(), exit(0);
		}
		process_reorder() ;
	}
	if(num<=10)
		print_adj_list();
	tottime= 0.0 ;
	if(mode == MODE_MAX) 
		tbestcost = VERY_LOW;
	else {
		tbestcost = VERY_HIGH;
		tbestminsize = 1;
	}
	if (alloc_b) {
		int maxdeg = find_max_degree();
		alloc_buckets(2*maxdeg+1);
		fprintf(outfile,"OUTPUT:\n");
		fprintf(outfile,"   edges:%6d\n",numedges);
		if(mode == MODE_MAX)
			fprintf(outfile,"Maximize number of edges cut\n");
		else
			fprintf(outfile,"Minimize number of edges cut\n");

		if(!output_short) {
			fprintf(outfile,"%-77s\n", hyphens);
			fprintf(outfile,"%-6s%-7s%-7s%-25s%-5s%-7s%-10s%-6s%-6s\n", 
				"PASS",  "ITER.",  "TIME",  "SOLUTION",  "DIFF",
				"GREEDY",  "NON-GR.", "SET_A","SET_B");
			fprintf(outfile,"%-77s\n", hyphens);
		}
	}
        hold_nongreedy = num-1;		/* several rounds of post	*/
					/* processing may be necessary	*/
					/* with the backtracking opt.	*/
					/* the most nongreedy is kept.	*/
}

void fix_verts(int begin_range, int end_range)
{
	int i;
	
	for (i=begin_range; i<end_range; i++) {
		nodes[vertices[i]].fix_next = TRUE;
	}
}

void save_and_postproc(int max_min, int begin_fix, int end_fix, 
		       int *bestcost,		/** for up_down KL **/
		       int *bestminsize)	/** for up_down KL **/
{
	extern int tbestcost;
	extern int tbestminsize;
	extern int Heuristic_Only;
	int i, temp_cost;
	double current_quotient=0.0, tbest_quotient=0.0, up_down_quotient=0.0;

	if (max_min == MIN_QUOTIENT) {
		current_quotient = (double)currentcost/(double)currentminsize;
		tbest_quotient   = (double)tbestcost  /(double)tbestminsize;
		up_down_quotient   = (double)*bestcost  /(double)*bestminsize;
	}
	if (((max_min == MODE_MAX) && (currentcost > tbestcost)) ||
            ((max_min == MODE_MIN) && (currentcost < tbestcost)) ||
            ((max_min == MIN_QUOTIENT) && (current_quotient < tbest_quotient))){
		tbestcost = currentcost;
		tbestminsize = currentminsize;
		hold2_num_vertices[SET_A] = num_vertices[SET_A];
		hold2_num_vertices[SET_B] = num_vertices[SET_B];
		for (i=0; i<num; i++)
			hold_kl_loc[i] = nodes[i].loc;
		temp_cost = currentcost;
		if (!Heuristic_Only)
			post_processing(TRUE, begin_fix, end_fix);
		currentcost = temp_cost;
	} else {
		temp_cost = currentcost;
		if (!Heuristic_Only)
			post_processing(FALSE, begin_fix, end_fix);
		currentcost = temp_cost;
	}
	/*** separate for up-down kl (different levels of "bestcost ***/
	if (((max_min == MODE_MAX) && (currentcost > *bestcost)) ||
            ((max_min == MODE_MIN) && (currentcost < *bestcost)) ||
            ((max_min == MIN_QUOTIENT) && (current_quotient<up_down_quotient))){
		*bestcost = currentcost;
		*bestminsize = currentminsize;
	}
}

void augment_kl_with_paths()
{ 
	extern int Heuristic_Only;
	/* extern int greedy, nongreedy;*/
	/* extern int num_vertices[2];*/

	time_ = log_time();
	currentcost=compute_costs();
	reset_buckets(node_cost);
	find_augmenting_paths();
	if (!Heuristic_Only && !check_cost(NULL, currentcost)) 
		printf("************************* program error #1a\n");
	time_ = log_time();

	tottime = tottime + time_;
/*	if (!output_short)
    		fprintf(outfile,"%45s%-7d%-10d%-6d%-6d\n", "",greedy,nongreedy,
                   		num_vertices[SET_A],num_vertices[SET_B]);*/
}

void print_iteration(int pass_num, int iter, 
		     int ubestcost, int ubestminsize,  
		     int diff)
{
	extern int Simulated_Annealing;
	extern double Current_Temp;
	extern int sort_type;
	int temp_greedy, temp_nongreedy, percent_ng;
        char   solution[20] = "                   ";

        if (Simulated_Annealing) {
        	if (mode == MIN_QUOTIENT)
               		 sprintf(solution,"%-6.4f %6.4f(%d)",
				  Current_Temp,
				  (double) ubestcost/(double) ubestminsize,
                                                ubestcost);
        	else
                	sprintf(solution,"%-10.4f %10d", Current_Temp, 
							 ubestcost);
	} else {
        	if (mode == MIN_QUOTIENT)
               		 sprintf(solution,"%-10.6f (%d)",
				  (double) ubestcost/(double) ubestminsize,
                                                ubestcost);
        	else
               		 sprintf(solution,"%-20d", ubestcost);
	}
	if (sort_type != BACKTRACK) {
		temp_greedy = greedy;
		temp_nongreedy = nongreedy;
	} else {
		temp_greedy = hold_greedy;
		temp_nongreedy = hold_nongreedy;
	}
	if (temp_greedy+temp_nongreedy > 0) 
		percent_ng = (temp_nongreedy*100)/ (temp_greedy+temp_nongreedy);
	else
		percent_ng = 0;

	   fprintf(outfile,"%-6d%-7d%-7.2f%-25s%-5d%-7d%-4d%3d%%%2s%-6d%-6d\n",
		pass_num,iter,time_,solution,
		diff,temp_greedy, temp_nongreedy,
		percent_ng,"", num_vertices[SET_A],num_vertices[SET_B]);
}

void w_start()
{
	extern int vert_loc[];  /** W sets this **/
	double unused;
	int i;

       	init_walk(RESET_LOCS);
       	one_walk(0,num, &unused);
	for (i=0; i<num; i++)
         	nodes[i].loc = vert_loc[i];
}

void kl_main_loop()
{
	extern int Produce_Random_Solution;
	extern int Augment_With_Paths;
	extern int Geometric_Graphs;
	extern int Line_Starts;
	extern int Up_Down_KL;
	extern int Up_Down_Reps;
	extern int Heuristic_Only;
	extern int Output_Greedy_History;
	extern int Done_With_Alg;	/* TRUE when all passes done    */
	extern int W_Starts;
	extern int seed;
	int hours=0;
	double last_time=0.0;
	int i=0, diff=0;
	int more_work=TRUE;
	int cur_best_nongreedy = num;
	int count = 0;
	int ubestcost;
	int ubestminsize;

	randinit(seed);		/* 4/27/93  - to compare with LEDA */

	if (Geometric_Graphs && Line_Starts)
		line_randomize();
	else
            	randomize_vertices();
	/*if (Output_Gnuplot)
	    	dump_cutfiles();*/

	while (more_work) {	
	    cur_best_nongreedy = num;
            currentcost = compute_initial_cost(node_cost, vertex_loc);

	    if (Output_Greedy_History)  
	       open_greedy_outfile(i);	/* keeps track of results of 	*/
					/* pretending each vertex is 	*/
					/* placed both ways		*/
	    if (!Produce_Random_Solution) {
              if(mode == MODE_MAX) {
		ubestcost = VERY_LOW;
		if (Up_Down_KL) {
               		perform_maximize(0, 0, &ubestcost);
			for (count=0; count<Up_Down_Reps; count++) {
				int lower = num/5;
				int upper = 4*num/5;
				cur_best_nongreedy = hold_nongreedy;
				if (count % 2  == 0) {
					fix_verts(0, lower);
               				perform_maximize(0, lower, &ubestcost);	
				} else {
					fix_verts(upper, num);
               				perform_maximize(upper,num, &ubestcost);
				}
			}
		} else
               		perform_maximize(0,0, &ubestcost);	
              } else {
		ubestcost = VERY_HIGH;
		ubestminsize = 1;
		if (Up_Down_KL) {
               		perform_minimize(0, 0, &ubestcost, &ubestminsize);
			for (count=0; count<Up_Down_Reps; count++) {
				int lower = num/5;
				int upper = 4*num/5;
				cur_best_nongreedy = hold_nongreedy;
				if (count % 2  == 0) {
					fix_verts(0, lower);
               				perform_minimize(0, lower, 
						&ubestcost, &ubestminsize);
				} else {
					fix_verts(upper, num);
               				perform_minimize(upper,num, 
						&ubestcost, &ubestminsize);
				}
			}
		} else
               		perform_minimize(0,0, &ubestcost, &ubestminsize);
	      }
	    } else {
		int dummy;
		if (mode == MODE_MAX)  
			dummy = VERY_LOW;
		else
			dummy = VERY_HIGH;
		ubestcost = currentcost;
		currentminsize = MIN(num_vertices[SET_A], num_vertices[SET_B]);
		save_and_postproc(mode, 0, 0, &dummy, &dummy);
	    }
	    /*if (Output_Gnuplot)
	    	dump_cutfiles();*/

	    if (Augment_With_Paths) {
		augment_kl_with_paths();
		ubestcost = currentcost;
		ubestminsize = currentminsize;
	    }

	    if (Output_Greedy_History) 
	       close_greedy_outfile();

	    /*diff = compare_and_save(i, hold_kl_loc);*/
            if(num <= 40) /** test mode **/
               print_locs(vertex_loc);
            if(!output_short) 
		print_iteration(i+1, iter, ubestcost, ubestminsize, diff);
	    i++;
	    if (Heuristic_Only) {
		if (tottime >= cpu_seconds)
			more_work = FALSE;
		if (tottime >= (last_time+3600.0)) {
			hours++;
			last_time = tottime;
			/*fprintf(tfile,"hours: %d, quot: %f\n", hours,
				(double) tbestcost/tbestminsize);
			fflush(tfile);*/
		}
	    } else {
		if (i >= num_passes)
			more_work = FALSE;
	    }
	    if (more_work) {
        	if (W_Starts) {
			w_start();
		} else if (Geometric_Graphs && Line_Starts)
			line_randomize();
	    	else
            		randomize_vertices();
	    }
	}
	/*if (Output_Gnuplot)
	    	dump_cutfiles(&vertices[(int) (num * 0.9)], (int) (num * 0.1));*/
	Num_Runs = i;
	Done_With_Alg = TRUE;
}
