/**************************************************************/
/*                     Graph Paritioning                      */
/*                            kl.c                            */
/*                                                            */
/*                   Author:  James P. Fox                    */
/*                   Advisor: Mark Goldberg                   */
/*                                                            */
/*                       Masters Project                      */
/*               Rensselaer Polytechnic Institute             */
/*                     Summer, Fall, 1990                     */
/**************************************************************/
/*     This program impliments the Kernighan & Lin algorithm  */
/*  partitioning. (The Bell System Technical Journal, Feb.,   */
/*  1970, pp. 291-307.)  This algorithm attempts to find the  */
/*  partitions which provide the maximum/minimum number of    */
/*  edges between the partitions.  The program accepts        */
/*  command options to perfrom the maximum or minimum number  */
/*  of cuts and can take input from a file or from standard   */
/*  input.  Output can be sent to the screen or optionally to */
/*  a disk file.  A command option is also provided to show a */
/*  graphed result of the greedy/non-greedy steps used to     */
/*  to create the partitions.                                 */
/*                                                            */
/*      The following input is required:                      */
/*         number of vertices             1..1000             */
/*         edge probability               0.0001..0.9999      */
/*         random number seed             0..2147483647       */
/*         number of applications of kl   1..99999            */
/*         post-processing sort type      0..8                */
/*         look-ahead                     0 - no              */
/*                                        1 - yes             */
/*         confirm process of this data   y/n                 */
/*                                                            */
/*      The following additionial input is required for       */
/*         different reordering methods:                      */
/*      5,6,8:                                                */
/*         expected ind. set size         1..1000             */
/*         limit of the bounds            0..999999           */
/*         number of bounds to enter      1..1000             */
/*         each individual bounds         1..1000             */
/*      7,8:                                                  */
/*         tie situation:                 1 - highest degree  */
/*                                        2 - lowest degree   */
/*                                                            */
/*      After input is received the graph is created and      */
/*  processed using the k-l algorithm.  After each            */
/*  application the best (min/max) partition information is   */
/*  printed including information on greedy/non-greedy steps. */
/*  After the last application of k-l the best partition is   */
/*  printed, along with an optional graph of greedy/non-      */
/*  greedy steps.                                             */
/**************************************************************/

#include <stdio.h>
#include <ctype.h>
#include "part.h"
#include "backtrack.h"

void init_backtrack(int);
void sim_annealing();
void free_geom_nodes();
void init_data_sets();
void init_clusters();
void read_data_sets();
void read_clusters();
void kl_init_loop(int);
void kl_main_loop();
void print_graphs();
void free_buckets();
void calc_ng_placed_neighbors();
void write_data_sets(int num_v);
void write_clusters(int num_v);
void process_parameters(int, char *argv[]);





void output_gp();
void output_ng();
void nongreedy_stats();

void get_config() {}	/* to avoid having to include use_config.c */

/*****************  input data  ******************************************/
extern int num ;                 /** number of vertices 		**/
extern double prob;               /** edge probability   		**/
/*************************************************************************/

extern int tbestcost;
extern int tbestminsize;
extern int hold2_num_vertices[2];	/** numbers in best partition  **/

extern FILE *outfile;
extern int end_of_file,process ;

extern int currentcost;   		/** maintain current cost info  **/
extern int currentminsize;   		/** of a partition 		**/
extern int output_short;  		/** output flags **/
extern int p_graph, mode;

extern int Geometric_Graphs;		/** true if using these, not (n,p) **/

extern double min_mult;
extern double t, tottime;

extern int Nongreedy_Stats;		/** flag - compare P_n(z), P_g(z)?**/
extern int NG_Upflag;			/** flag - make upneigh. file (-u)**/

extern FILE *read_sng_file();
extern FILE *read_ng_upfile();
extern void write_sng_file(int); 
extern void write_ng_upfile(int); 
extern FILE *Sng_Distfile;
extern FILE *NG_Upfile;
extern int alg_type;			/** KL, BACKTRACK */ 
int backup_num;				/* num goes to 0 when done */
					/* remember what num was.  */
int num_graphs;				/* how many in this run?   */


void main(int argc, char *argv[]) 
{  
	extern int Geometric_Graphs;
	extern int Output_Sng_Distfile;
	extern int Compute_Variance;
	extern int Find_Clusters;
	extern int Simulated_Annealing;
	extern int W_Starts;
	extern double time_;
	extern STK_ELEMPTR stack;


	int first=TRUE;

	num_graphs=0;			
	alg_type = ALG_TYPE_KL;
	clear_summary();
	process_parameters(argc,argv);
	end_of_file = FALSE;

	if (Output_Sng_Distfile || NG_Upflag)
		init_extra_file_info();
	if (Compute_Variance)
		init_data_sets();
	if (Find_Clusters)
		init_clusters();

	while(!end_of_file) {
	      if(!read_data()) {
			printf("invalid data input, program terminated....\n");
			exit(1);
	      }
	      if (first) {
			if (Output_Sng_Distfile) {
				Sng_Distfile = read_sng_file();
			} if (NG_Upflag) 
				NG_Upfile = read_ng_upfile();
			if (Compute_Variance)
				read_data_sets();
			if (Find_Clusters)
				read_clusters();
			first = FALSE;
	      }
	      if (process) {
		kl_init_loop(ALLOC);
        	if (W_Starts)
               		init_backtrack(DONT_CREATE_GRAPH);

		if (Simulated_Annealing)
			sim_annealing();
		else
			kl_main_loop();

        	if (W_Starts)
			free(stack);

		if(output_short)
		     fprintf(outfile,"Cuts in best partition: %5d\n",tbestcost);
		else {
			char   solution[20];

			if (MIN_QUOTIENT)
				sprintf(solution,"%10.6f (%d)",
					(double) tbestcost/(double)tbestminsize,
                               		tbestcost);
			else
				sprintf(solution,"%-20d", tbestcost);
		      	fprintf(outfile,"\n TOTAL->-> %6.2f%-20s\n", 
				tottime, solution);
		}
		time_ = tottime;
		currentcost = tbestcost;
		currentminsize = tbestminsize;
		if (Nongreedy_Stats)
			nongreedy_stats();
		add_to_summary();
		if(p_graph)
			print_graphs();
		fflush(outfile);
		free_adj_list() ;
		free_buckets();
		if (Geometric_Graphs)
			free_geom_nodes();
		backup_num = num;
		num_graphs++;
		calc_ng_placed_neighbors();	/* for ng_upfile */
		if (num_graphs % 10 == 0) {
			if (Output_Sng_Distfile)
				write_sng_file(num);
			if (NG_Upflag)
				write_ng_upfile(num);
			if (Compute_Variance)
				write_data_sets(num);
			if (Find_Clusters)
				write_clusters(num);
		}
	      }
	}
	if (Output_Sng_Distfile)
		write_sng_file(backup_num);
	if (NG_Upflag)
		write_ng_upfile(backup_num);
	if (Compute_Variance)
		write_data_sets(backup_num);
	if (Find_Clusters)
		write_clusters(backup_num);
	num = 0; prob = 0.0; /** force final summary **/
	print_summary();
	exit(0);
}
